﻿/*************************************************************************************************************************************
 * FFT/IFFTクラス及び、処理結果を操作するクラス
 *  開発者：Katsuhiro Morishita(2010.5) Kumamoto-University
 * 
 * [名前空間]   SignalProcess
 * [クラス名]   FFTresult, DFT
 * 
 * 
 * [使用方法]   まず、プロジェクトに本ファイルを追加した後に、ソースヘッダに以下のコードを追加して下さい。
 *              using SignalProcess;
 *              
 * [更新情報]   2010/5/23   森下整備開始 （関係ないけど、初めてのC#だったりする）
 *              2011/2/18   名前空間を名づけた
 *                          今後：渡されたデータ幅がFFTデータ幅を超えた場合でも、可能なだけ結果を返すようにしたい。
 *              2011/2/26   コンストラクタでのエラー対策と、コメントを入れる。
 *                          拡張予定：　帯域指定でのパワー計算値を返すメソッド（引数にサンプリング周波数情報が必須）
 *                          窓関数に三角窓も追加予定
 *              2011/9/30   コメントの見直しを実施
 *              2011/11/15  コメントの見直しを更に実施
 *                          関数の一部整理
 *                          サンプリング周波数に関してメソッドを追加
 *                          FFT後のデータを取り扱いやすくするため、FFTresultクラスを整備
 *                          FFTdata構造体を少し改良
 *              2011/12/6   Windowing()メソッドで、矩形窓を指定した場合に全部0になるバグがあったのを修正
 *                          名前空間をSignal_ProcessからSignalProcessに変更
 *                          Windowingは他の処理でも使うことがあるので、同じ名前空間にWindowクラスとして独立させた。
 *                          それに伴い、ファイルを分割したので注意のこと。
 *                          Band構造体のコンストラクタ生成時に、引数の異常を検出して例外をスローするように変更
 *                          窓関数付の帯域パワー取得メソッド GetPSD(Band band, Window.WindowKind windowKind) を新設
 *              2011/12/8   コメントの一部見直し
 *                          IFFTをするポテンシャルはあるのに、実施するためのメソッドが用意されていないことに気が付いた。必要があれば作る予定。
 *                          FFTdataに位相情報を追加
 *              2012/1/17   FFTウィンドウ幅を最小32になるように調整した。
 *                          以前は条件によるものの、最小で2とされエラーがスローされることがあった。。。
 *                          帯域を指定してパワーを得られる、GetPowerArray(Band band)とGetFilterdResult(Band band)を新設。
 *              2012/1/18   周波数の取り扱いをdoubleに変更
 * *********************************************************************************************************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/******************************************
 * 信号処理を行うクラスを集めた名前空間です。
 * ****************************************/
namespace SignalProcess
{
    /********************************************************
     * FFT/IFFT関連クラス
     *  開発者：Katsuhiro Morishita(2011.2) Kumamoto-University
     *  
     *  参考文献:ニューメリカル・レシピ・イン・シー
     *           用語はこの書籍に準ずる。
     *  更新履歴：2010/5/23   森下整備開始
     * ******************************************************/
    #region "FFT/IFFT関連クラス"
    /// <summary>
    /// FFTデータ幅
    /// データサイズ(2^n)をこの列挙体を使って宣言すれば間違えない.
    /// </summary>
    public enum FFTpoint :int{
        /// <summary>データ幅32ポイント</summary>
        size32 = 32,
        /// <summary>データ幅64ポイント</summary>
        size64 = 64,
        /// <summary>データ幅128ポイント</summary>
        size128 = 128,
        /// <summary>データ幅256ポイント</summary>
        size256 = 256,
        /// <summary>データ幅512ポイント</summary>
        size512 = 512,
        /// <summary>データ幅1024ポイント</summary>
        size1024 = 1024,
        /// <summary>データ幅2048ポイント</summary>
        size2048 = 2048,
        /// <summary>データ幅4096ポイント</summary>
        size4096 = 4096,
        /// <summary>データ幅8192ポイント</summary>
        size8192 = 8192
    }
    /// <summary>
    /// FFT処理前/後のデータを格納することを想定した構造体
    /// </summary>
    public struct FFTdata
    {
        /************** メンバ変数 *********************/
        /// <summary>実部</summary>
        public double Re;
        /// <summary>虚部</summary>
        public double Im;
        /// <summary>位相</summary>
        public double Phase;
        /// <summary>絶対値</summary>
        public double Abs;
        /// <summary>絶対値の対数</summary>
        public double Log;
        /// <summary>パワー<para>ところで、パワースペクトル密度 （Power Spectrum Density,PSD）は、Powerを周波数分解能で割れば求まります。</para></summary>
        public double Power;                            // FFTの結果を格納した場合、最小の分解能の周波数幅におけるパワーの積算値でもある
        /************** メソッド *********************/
        /// <summary>
        /// データをセットする
        /// </summary>
        /// <param name="_Re">実部</param>
        /// <param name="_Im">虚部</param>
        public void Set(double _Re, double _Im)
        {
            this.Re = _Re;
            this.Im = _Im;
            this.Phase = Math.Atan2(this.Im, this.Re);
            this.Abs = Math.Sqrt(Math.Pow(_Re, 2) + Math.Pow(_Im, 2));
            this.Log = Math.Log10(this.Abs);
            this.Power = Math.Pow(_Re, 2) + Math.Pow(_Im, 2);
        }
    }
    /// <summary>
    /// 周波数帯域を定義する構造体
    /// </summary>
    public struct Band
    {
        /************ メンバ変数 ****************/
        /// <summary>最大周波数[Hz]</summary>
        public readonly double Max;
        /// <summary>最小周波数[Hz]</summary>
        public readonly double Min;
        /************ プロパティ ****************/
        /// <summary>
        /// バンド幅[Hz]
        /// </summary>
        public double BandWidth { get { return this.Max - this.Min; } }
        /// <summary>
        /// 中心周波数[Hz]
        /// </summary>
        public double CenterFrequency { get { return (this.Max - this.Min) / 2 + this.Min; } }
        /************ メソッド ******************/
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="max">最大周波数[Hz]</param>
        /// <param name="min">最小周波数[Hz]</param>
        /// <exception cref="SystemException">引数の大小関係が異常である場合にスローします。</exception>
        public Band(double max, double min)
        {
            if (max > min && min >= 0)
            {
                this.Max = max;
                this.Min = min;
            }
            else
            {
                throw new SystemException("引数の大小関係が異常です。");
            }
            return;
        }
    }
    /// <summary>
    /// FFT処理後の観測データを扱うクラス
    /// </summary>
    public partial class FFTresult
    {
        /**************************** 構造体 *******************************/
        /// <summary>
        /// FFTの処理データと周波数をセットにした構造体
        /// </summary>
        public struct FFTresultPlusFreq
        {
            /// <summary>FFT処理後のデータ</summary>
            public readonly FFTdata data;
            /// <summary>周波数[Hz]</summary>
            public readonly double frequency;
            /// <summary>
            /// コンストラクタ
            /// </summary>
            /// <param name="x">FFT処理後のデータ</param>
            /// <param name="freq">周波数[Hz]</param>
            public FFTresultPlusFreq(FFTdata x, double freq)
            {
                this.data = x;
                this.frequency = freq;
            }
        }
    }
    /// <summary>
    /// FFT処理後の観測データを扱うクラス
    /// <para>クラスの本体です。</para>
    /// </summary>
    public partial class FFTresult
    {
        /**************************** グローバル変数 *******************************/
        /// <summary>FFT処理データ</summary>
        private FFTdata[] _data;
        /// <summary>サンプリング周波数[Hz]</summary>
        private double _samplingFrequency = 0;
        /// <summary>FFTのサイズ（例：4096 point）</summary>
        private int _fftPoint = 0;
        /**************************** プロパティ *******************************/
        /// <summary>
        /// 観測データ数
        /// </summary>
        public int Length { get { return this._data.Length / 2; } }
        /// <summary>
        /// 表現されている最大の周波数[Hz]
        /// </summary>
        public double MaxFrequency { get { return this._samplingFrequency / 2L; } }
        /// <summary>
        /// 表現されている最小の周波数分解能[Hz]
        /// </summary>
        public double FrequencyResolution { get { return this.GetFreq(1); } }
        /**************************** メソッド *******************************/
        /// <summary>
        /// 結果を文字列で出力する
        /// </summary>
        /// <returns>文字列で表した結果</returns>
        public override string ToString()
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder(5000);
            sb.Append("Frequency[Hz]").Append(",").Append("Altitude").Append("\n");
            for(int i = 0; i < this.Length; i++)
                sb.Append(this.GetFreq(i)).Append(",").Append(this._data[i].Log).Append("\n");
            return sb.ToString();
        }
        /// <summary>
        /// 結果をファイルに保存する
        /// </summary>
        /// <param name="fname">ファイル名</param>
        public void Save(string fname = "hoge.csv")
        {
            using (System.IO.StreamWriter sw = new System.IO.StreamWriter(fname, false, System.Text.Encoding.GetEncoding("shift_jis")))
            {
                sw.Write(this.ToString());
            }
            return;
        }
        /// <summary>
        /// 要素番号を入れると、その要素の周波数を計算
        /// </summary>
        /// <param name="index">要素番号</param>
        /// <returns>周波数[Hz]</returns>
        private double GetFreq(int index)
        {
            return (double)index * this._samplingFrequency / ((double)this._fftPoint - 2);
        }
        /// <summary>
        /// 指定した要素番号のFFTの結果を取得
        /// </summary>
        /// <param name="index">要素番号（0～this.Lenght - 1）</param>
        /// <returns>周波数と結びつけたFFT結果</returns>
        public FFTresultPlusFreq GetData(int index)
        {
            if (index < this._data.Length && index >= 0)                         // 要素番号のチェック
                return new FFTresultPlusFreq(this._data[index], this.GetFreq(index));
            else
                return new FFTresultPlusFreq();
        }
        /// <summary>
        /// パワーをdouble型配列として返す
        /// </summary>
        /// <returns>double型一次元配列に格納されたパワー</returns>
        public double[] GetPowerArray()
        {
            double[] ans = new double[this.Length];
            for (int i = 0; i < ans.Length; i++) ans[i] = this._data[i].Power;
            return ans;
        }
        /// <summary>
        /// パワーに指定帯域の矩形窓をかけて、double型配列として返す
        /// <para>指定した帯域のみ値を格納するし、指定帯域外は0.0を格納します。</para>
        /// <para>配列には折り返し以降の周波数は含まれません。</para>
        /// </summary>
        /// <param name="band">帯域</param>
        /// <returns></returns>
        public double[] GetPowerArray(Band band)
        {
            double[] ans = new double[this.Length];
            for (int i = 0; i < ans.Length; i++)
            {
                double freq = this.GetFreq(i);
                if (freq >= band.Min && freq <= band.Max)
                    ans[i] = this._data[i].Power;
                else
                    ans[i] = 0.0;
            }
            return ans;
        }
        /// <summary>
        /// 指定帯域でフィルタをかけたFFTresultオブジェクトを返す
        /// </summary>
        /// <param name="band">帯域</param>
        /// <returns>フィルタリングされたFFTresult<para>サンプリング周波数より先の折り返し分はコピーされません。</para></returns>
        public FFTresult GetFilterdResult(Band band)
        { 
            FFTdata[] copy = new FFTdata[this._data.Length];
            for (int i = 0; i < copy.Length; i++)
            {
                double freq = this.GetFreq(i);
                if (freq >= band.Min && freq <= band.Max)
                    copy[i] = this._data[i];
                else
                    copy[i] = new FFTdata();
            }
            return new FFTresult(copy, this._samplingFrequency, this._fftPoint);
        }
        /// <summary>
        /// 指定周波数がどの要素番号に該当するかを返す
        /// </summary>
        /// <param name="frequency">周波数[Hz]</param>
        /// <returns>要素番号</returns>
        private int GetIndex(double frequency)
        {
            int index = 0;
            if (this._samplingFrequency != 0 && this._fftPoint != 0)
                index = (int)(frequency * ((long)this._fftPoint - 2) / this._samplingFrequency);
            return index;
        }
        /// <summary>
        /// 指定帯域におけるパワースペクトル密度 （Power Spectrum Density,PSD）を返す
        /// <para>例えば、20kHzまでしか帯域パワー場存在しない（@サンプリング周波数が40kHz）場合、30kHzまでの帯域パワー密度を要求すると20kHzまでを指定した時と比較して小さな値となります。</para>
        /// </summary>
        /// <param name="min_frequency">帯域の下限周波数[Hz]</param>
        /// <param name="max_frequency">帯域の上限周波数[Hz]</param>
        /// <returns>パワースペクトル密度[V^2/Hz]</returns>
        /// <exception cref="SystemException">帯域パワーを定義できない場合にスロー</exception>
        public double GetPSD(double min_frequency, double max_frequency)
        {
            double psd = 0.0;

            if (this._samplingFrequency == 0 || this._fftPoint == 0) // エラーを防ぐための処置 
            { 
                Console.WriteLine("FFTresultクラスのGetPSDメソッドにてエラーがスローされました。\n帯域パワーを定義できませんでした。");
                throw new SystemException("帯域パワーを定義できませんでした。");
            }
            // 指定された帯域幅におけるインデックスの幅をあらかじめ計算しておく
            int indexWidth = this.GetIndex(max_frequency) - this.GetIndex(min_frequency) + 1;
            // 指定された帯域があり得なければ修正
            if (min_frequency < 0) min_frequency = 0;
            if (max_frequency > (this._samplingFrequency / 2)) max_frequency = (this._samplingFrequency / 2);
            // 可能なら帯域パワーを計算する
            if (min_frequency < max_frequency)
            {
                int max_index = this.GetIndex(max_frequency);           // 最大周波数に対応した要素番号を取得
                int min_index = this.GetIndex(min_frequency);           // 最小周波数に対応した要素番号を取得
                for (int i = min_index; i <= max_index; i++)
                    psd += this.GetData(i).data.Power;                  // 積分
                psd /= ((double)indexWidth * (double)this.FrequencyResolution); // 単位周波数あたりに変換
            }
            else
            {
                psd = 0.0;
            }
            return psd;
        }
        /// <summary>
        /// 指定帯域におけるパワースペクトル密度 （Power Spectrum Density,PSD）を返す その2
        /// <para>帯域を指定できる構造体を使用しているので、配列での処理がやりやすくなったと思う。</para>
        /// </summary>
        /// <param name="band">帯域</param>
        /// <returns>パワースペクトル密度[V^2/Hz]</returns>
        public double GetPSD(Band band)
        {
            return this.GetPSD(band.Min, band.Max);
        }
        /// <summary>
        /// 指定帯域におけるパワースペクトル密度 （Power Spectrum Density,PSD）を返す その3
        /// <para>窓関数を掛けた帯域パワーを返します。</para>
        /// </summary>
        /// <param name="band">帯域</param>
        /// <param name="windowKind">窓関数の種類</param>
        /// <returns>パワースペクトル密度[V^2/Hz]</returns>
        /// <exception cref="SystemException">帯域パワーを定義できない場合にスロー</exception>
        public double GetPSD(Band band, Window.WindowKind windowKind)
        {
            double psd = 0.0;

            if (this._samplingFrequency != 0 && this._fftPoint != 0 && band.Max < (this._samplingFrequency / 2)) // エラーを防ぐための処置 
            {
                int max_index = this.GetIndex(band.Max);                // 最大周波数に対応した要素番号を取得
                int min_index = this.GetIndex(band.Min);                // 最小周波数に対応した要素番号を取得
                double[] psd_arr = new double[max_index - min_index + 1];
                for (int i = min_index, k = 0; i <= max_index; i++, k++)
                    psd_arr[k] = this.GetData(i).data.Power;            // 指定帯域のパワーを取得
                psd_arr = Window.Windowing(psd_arr, windowKind);        // 窓関数を掛ける
                foreach (double v in psd_arr) psd += v;                 // 積分
                psd /= ((double)(max_index - min_index + 1) * (double)this.FrequencyResolution); // 単位周波数あたりに変換
            }
            else throw new SystemException("帯域パワーを定義できませんでした。");
            return psd;
        }
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="setData">観測データ配列</param>
        /// <param name="samplingFrequency">元データのサンプリング周波数</param>
        /// <param name="FFTpoint">FFTポイント数</param>
        public FFTresult(FFTdata[] setData, double samplingFrequency, int FFTpoint)
        { 
            this._data = new FFTdata[setData.Length];
            for (int i = 0; i < setData.Length; i++) this._data[i] = setData[i]; // 副作用がないように、一つずつコピー
            this._samplingFrequency = samplingFrequency;
            this._fftPoint = FFTpoint;
        }
    }
    /// <summary>
    /// <para>FFT若しくはIFFTを実施するためのクラスです。</para>
    /// </summary>
    /// <example>
    /// DFTクラスの使用例です。
    /// <code>
    /// DFT fft = new DFT[4096];
    /// 
    /// FFTresult result = fft_arr[i].FFT(sound, 44100); // soundはdouble型又はFFTdata型の一次元配列で、処理したいデータが入っている
    /// </code>
    /// </example>
    public class DFT
    {
        /***********************　列挙体の宣言  *******************************/
        /// <summary>
        /// FFT / IFFT
        /// </summary>
        private enum FFTorIFFT : int { 
            FFTconvert = 1,
            IFFTconvert = -1
        }
        /***********************　グローバル変数の宣言  *******************************/
        /// <summary>FFTデータ幅</summary>
        private int __FFTpoint;
        /// <summary>使用する窓の種類</summary>
        private Window.WindowKind __WindowKind;
        /// <summary>計算後の配列</summary>
        private FFTdata[] calced_datas;
        /// <summary>外部より渡されたデータを格納する配列</summary>
        private FFTdata[] set_datas;
        /// <summary>データのサンプリング周波数</summary>
        private double frequencyOfSample = 0.0;
        /****************************　プロパティ　************************************/
        /// <summary>
        /// 窓関数の種別
        /// </summary>
        public Window.WindowKind WindowKind
        {
            get { return this.__WindowKind; }
            set { this.__WindowKind = value; }
        }
        /// <summary>
        /// FFT/IFFTデータ幅
        /// </summary>
        public int FFTsize { get { return this.__FFTpoint; } }
        /// <summary>
        ///  変換後の配列に含まれる最大値を返す
        /// </summary>
        public double Max
        {
            get
            {
                double max = 0;
                int i;

                for (i = 0; i < this.calced_datas.Length; i++) {
                    if (max < this.calced_datas[i].Abs) max = this.calced_datas[i].Abs;
                }
                return max;
            }
        }
        /// <summary>
        /// 変換後の配列に含まれる平均値を返す
        /// </summary>
        public double Mean
        {
            get
            {
                double mean = 0;
                int i;

                for (i = 0; i < this.calced_datas.Length; i++)
                {
                    mean += this.calced_datas[i].Abs;
                }
                return mean / this.calced_datas.Length;
            }
        }

        /****************************　メソッド　************************************/
        /// <summary>
        /// データをセットするメソッド。
        /// <para>
        /// 指定された窓関数をかけて保持する。
        /// データサイズよりFFTサイズの方が大きい場合は、0で埋める。
        /// 反対に、小さい場合は入るだけしか受け取らない。
        /// </para>
        /// </summary>
        private void Dataset(FFTdata[] arr) {
            int i;
            int size;
            double[] temp = new double[arr.Length];

            size = arr.Length;
            if (this.__FFTpoint < arr.Length) size = this.__FFTpoint;   // フィルタにかけるサイズを決定する
            // Reパート
            for (i = 0; i < arr.Length; i++) temp[i] = arr[i].Re;
            temp = Window.Windowing(temp, this.__WindowKind);           // 窓を掛ける
            for (i = 0; i < this.__FFTpoint; i++)                       // 受け渡されたデータをコピー
            {
                if( arr.Length > i )                                    // ある限りコピー
                    this.set_datas[i].Re = temp[i];
                else
                    this.set_datas[i].Re = 0.0;                         // 余ったら0を詰める
            }

            // Imパート
            for (i = 0; i < arr.Length; i++) temp[i] = arr[i].Im;
            temp = Window.Windowing(temp, this.__WindowKind);           // 窓を掛ける
            for (i = 0; i < this.__FFTpoint; i++)                       // 受け渡されたデータをコピー
            {
                if (arr.Length > i)
                    this.set_datas[i].Im = temp[i];                     // ある限りコピー
                else
                    this.set_datas[i].Im = 0.0;                         // 余ったら0を詰める
            }
            return;
        } 
        /// <summary>
        /// データをセットするメソッド2
        /// <para>
        /// 指定された窓関数をかけて保持する。
        /// データサイズよりFFTサイズの方が大きい場合は、0で埋める。
        /// </para>
        /// </summary>
        private void Dataset(double[] arr) {
            FFTdata[] copy = new FFTdata[arr.Length];

            for (int k = 0; k < arr.Length; k++)
            { 
                copy[k].Re = arr[k];
                copy[k].Im = 0.0;
            }
            this.Dataset(copy);
            return;
        }
        /// <summary>
        /// FFT/IFFTメソッド
        /// </summary>
        /// <param name="data">
        /// 被変換変数。結果が格納される。長さnnの複素配列（実部と虚部を一列に並べたもの）。
        /// 配列の中には、実部・虚部・実部・虚部・実部・虚部・実部・虚部・…と並んでおればよい。
        /// 原本では要素番号1から詰めていることを仮定している。
        /// 実部と虚部が入れ替わっても特に影響は無いはず。
        /// </param>
        /// <param name="nn">data[]に格納されたデータ長を表す。要素数の半分である。2の整数乗である必要がある。</param>
        /// <param name="isign">isign　 1：FFT変換を行う　-1：IFFT変換を行う。結果はnn倍される。</param>
        private void Calc(double[] data, long nn , int isign ){
            long n, mmax, m, j, istep, i;
            double wtemp, wr, wpr, wpi, wi, theta;              // 三角関数の漸化式用
            double tempr, tempi;

            n = nn << 1;                                        // n:データポインタ数の2倍
            j = 1;
            for (i = 1; i < n; i += 2) {                        // ビット反転アルゴリズム
                if (j > i) {                                    // 複素数を交換
                    tempr = data[j]; data[j] = data[i]; data[i] = tempr;                    // 原本では実部を表す
                    tempi = data[j + 1]; data[j + 1] = data[i + 1]; data[i + 1] = tempi;    // 原本では虚部を表す
                }
                m = n >> 1;
                while (m >= 2 && j > m) {
                    j -= m;
                    m >>= 1;
                }
                j += m;
            }
            mmax = 2;                                           // 以下はDanielson-Lanczosアルゴリズムを採用
            while (n > mmax) {                                  // 外側のループはlog2 nn回実行される
                istep = mmax << 1;
                theta = isign * (6.28318530717959 / mmax);      // 三角関数の漸化式の初期値. 2*pi = 6.28…
                wtemp = Math.Sin (0.5 * theta);
                wpr = -2.0 * wtemp * wtemp;
                wpi = Math.Sin(theta);
                wr = 1.0;
                wi = 0.0;
                for (m = 1; m < mmax; m += 2) {
                    for (i = m; i <= n; i += istep) {
                        j = i + mmax;                           // Danielson-Lanczos公式
                        tempr = wr * data[j] - wi * data[j + 1];// 原本通りならreal part
                        tempi = wr * data[j + 1] + wi * data[j];// 原本通りならimaginary part
                        data[j] = data[i] - tempr;
                        data[j + 1] = data[i + 1] - tempi;
                        data[i] += tempr;
                        data[i+1] += tempi;
                    }                                           // 三角関数の漸化式
                    wr = (wtemp = wr) * wpr - wi * wpi + wr;    // ここは処理をひっくり返せばwtempをそのままwrに置き換えられるが、恐らくrealパートから計算したいのだろう。
                    wi = wi * wpr + wtemp * wpi + wi;
                }
                mmax = istep;
            }
            return;
        }
        #region "FFTの実行インターフェイス"
        /// <summary>
        /// double型一次元配列に対して、FFTを実行する
        /// </summary>
        /// <example>
        /// <code>
        /// DFT a = new DFT();                              // このクラスを宣言して、実態も確保
        /// FFTresult result = a.FFT(temp);                 // double型配列を渡してFFT演算を実行。配列サイズがFFTデータ幅とする必要はない。
        /// </code>
        /// </example>
        /// <param name="data">double型の配列に解析したいデータを格納してください</param>
        /// <returns>FFT処理データ</returns>
        public FFTresult FFT(double[] data)
        {
            int i,k;
            double[] data_for_calc = new double[this.__FFTpoint * 2 + 1];

            this.Dataset(data);                                                     // データをセットする
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                data_for_calc[k] = this.set_datas[i].Re;                            // 解析するデータを格納する
                data_for_calc[k + 1] = 0.0;                                         // こっちはimaginaryパート
            }
            this.Calc(data_for_calc, this.__FFTpoint, (int)FFTorIFFT.FFTconvert);   // FFT演算
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                this.calced_datas[i].Set(data_for_calc[k], data_for_calc[k + 1]);   // 結果を格納する
            }
            return new FFTresult(this.calced_datas, this.frequencyOfSample, this.__FFTpoint);
        }
        /// <summary>
        /// double型一次元配列に対して、FFTを実行する
        /// <para>サンプリング周期をセットする機能付き（結果の整理が楽になります）</para>
        /// </summary>
        /// <param name="data">解析したいデータ配列</param>
        /// <param name="frequency">サンプリング周波数[Hz]</param>
        /// <returns>FFT処理データ</returns>
        public FFTresult FFT(double[] data, double frequency)
        {
            this.frequencyOfSample = frequency;
            return this.FFT(data);
        }
        /// <summary>
        /// FFTdata型一次元配列に対して、FFTを実行する
        /// <para>実部・虚部を合わせて解析するFFTメソッド</para>
        /// </summary>
        /// <param name="data">FFTdatas型の配列に解析したいデータを格納してください</param>
        /// <returns>FFT処理データ</returns>
        public FFTresult FFT(FFTdata[] data) 
        {
            int i, k;
            double[] data_for_calc = new double[this.__FFTpoint * 2 + 1];

            this.Dataset(data);                                                     // データをセットする
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                data_for_calc[k] = this.set_datas[i].Re;                            // 解析するデータを格納する
                data_for_calc[k + 1] = this.set_datas[i].Im;                        // こっちはimaginaryパート
            }
            this.Calc(data_for_calc, this.__FFTpoint, (int)FFTorIFFT.FFTconvert);   // FFT演算
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                this.calced_datas[i].Set(data_for_calc[k], data_for_calc[k + 1]);   // 結果を格納する
            }
            return new FFTresult(this.calced_datas, this.frequencyOfSample, this.__FFTpoint);
        }
        /// <summary>
        /// FFTdata型一次元配列に対して、FFTを実行する
        /// <para>実部・虚部を合わせて解析するFFT/IFFTメソッド</para>
        /// <para>サンプリング周期をセットする機能付き（結果の整理が楽になります）</para>
        /// </summary>
        /// <param name="data">解析したいデータ配列</param>
        /// <param name="frequency">サンプリング周波数[Hz]</param>
        /// <returns>FFT処理データ</returns>
        public FFTresult FFT(FFTdata[] data, double frequency)
        {
            this.frequencyOfSample = frequency;
            return this.FFT(data);
        }
        #endregion
        /****************************　初期化/削除　************************************/
        /*****************
         * このクラスのコンストラクタ
         * 
         * 初期化時に使用します。
         * ***************/
        /// <summary>
        /// 　デフォルトでは4096ポイントFFT/IFFT，窓関数の種別としてハミングウィンドウがセットされます。
        /// </summary>
        /// <param name="FFT_point">FFT/IFFTポイント数</param>
        /// <param name="window_kind">窓関数の種類</param>
        public DFT(FFTpoint FFT_point = FFTpoint.size4096, Window.WindowKind window_kind = Window.WindowKind.Hanning)
        {
            this.__FFTpoint = (int)FFT_point;
            this.__WindowKind = window_kind;

            this.calced_datas = new FFTdata[this.__FFTpoint];       // サイズに合わせて、インスタンスを生成
            this.set_datas = new FFTdata[this.__FFTpoint];
        }
        /// <summary>
        /// 　デフォルトでは4096ポイントFFT/IFFT，窓関数の種別としてハミングウィンドウがセットされます。
        /// 　指定ポイント数が2^Nではない場合、FFT_point＜2^N かつ32以上となる最小の2^Nを採用しますので注意.
        /// </summary>
        /// <param name="FFT_point">FFT/IFFTポイント数</param>
        /// <param name="window_kind">窓関数の種類</param>
        public DFT(int FFT_point = 4096, Window.WindowKind window_kind = Window.WindowKind.Hanning)
        {
            int FFT_point2 = 1;

            while (FFT_point2 < FFT_point || FFT_point2 < (int)FFTpoint.size32) FFT_point2 <<= 1;        // サイズ調整
            this.__FFTpoint = FFT_point2;
            this.__WindowKind = window_kind;
            
            this.calced_datas = new FFTdata[this.__FFTpoint];       // サイズに合わせて、インスタンスを生成
            this.set_datas = new FFTdata[this.__FFTpoint];
        }
    }
    #endregion
}
